# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.config import DopplerrConfig

log = logging.getLogger(__name__)


class _FilterBase(object):
    def appy_path_mapping(self, folder):
        if not DopplerrConfig().get_cfg_value("general.mapping"):
            return folder
        if folder.startswith("/"):
            absolute = True
            folder = folder[1:]
        for mapping in DopplerrConfig().get_cfg_value("general.mapping"):
            log.debug("Mapping: %s", mapping)
            k, _, v = mapping.partition("=")
            log.debug("Applying mapping %s => %s", k, v)
            if folder.startswith(k):
                folder = v + folder[len(k):]
                break
        if absolute:
            return "/" + folder
        return folder

    @staticmethod
    def lowerize_dick_keys(thedict):
        return {k.lower(): v for k, v in thedict.items()}


#
# class AutoDetectFilter(_FilterBase):
#     def filter(self, request, res):
#         low_request = self.lowerize_dick_keys(request)
#         if "series" in low_request:
#             return SonarrFilter().filter(request, res)
#         return res.failed("Unable to find request type")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

log = logging.getLogger(__name__)


class _NotificationBase(object):

    NOTIFICATION_TYPES = ["fetched"]

    def __init__(self, registered_notification):
        self._registered_notification = registered_notification

    def can_emit_notification_type(self, requested_notif_type):
        return requested_notif_type in self._registered_notification



# -*- coding: utf-8 -*-
"""
test_dopplerr
--------------
Tests for `dopplerr` module.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from dopplerr.downloader import DopplerrDownloader


class Testdopplerr(object):
    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test_sonarr_notify(self):
        pass

    def test_glob_filename_with_bracket(self):
        downloader = DopplerrDownloader()
        downloader.search_file("/any/root/dir", "base_name with [bracket]")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from pathlib import Path

from dopplerr.db import DopplerrDb
from dopplerr.downloader import DopplerrDownloader

log = logging.getLogger(__name__)


async def download_missing_subtitles(res):
    candidates = res.candidates
    if not candidates:
        DopplerrDb().insert_event("error", "event handled but no candidate found")
        log.debug("event handled but no candidate found")
        res.failed("event handled but no candidate found")
        return res

    for candidate in candidates:
        log.info(
            "Searching episode '%s' from series '%s'. Filename: %s",
            candidate.get("episode_title"),
            candidate.get("series_title"),
            candidate.get("scenename"),
        )
        DopplerrDb().insert_event("availability", "Available: {} - {}x{} - {} [{}].".format(
            candidate.get("series_title"),
            candidate.get("season_number"),
            candidate.get("episode_number"),
            candidate.get("episode_title"),
            candidate.get("quality"),
        ))

        video_files_found = DopplerrDownloader().search_file(candidate['root_dir'],
                                                             candidate['scenename'])
        log.debug("All found files: %r", video_files_found)
        if not video_files_found:
            res.failed("candidates found but no video file found")
            DopplerrDb().insert_event("subtitles", "No video file found for sonarr notification")
            return res
        DopplerrDb().update_series_media(
            series_title=candidate.get("series_title"),
            tv_db_id=candidate.get("tv_db_id"),
            season_number=candidate.get("season_number"),
            episode_number=candidate.get("episode_number"),
            episode_title=candidate.get("episode_title"),
            quality=candidate.get("quality"),
            video_languages=None,
            media_filename=video_files_found[0],
            dirty=True)

        await DopplerrDownloader().download_missing_subtitles(res, video_files_found)
        subtitles = res.subtitles
        if not subtitles:
            DopplerrDb().insert_event("subtitles", "no subtitle found for: {}".format(
                ", ".join([Path(f).name for f in video_files_found])))
            return res
        DopplerrDb().insert_event("subtitles", "subtitles fetched: {}".format(
            ", ".join([
                "{} (lang: {}, source: {})".format(
                    s.get("filename"),
                    s.get("language"),
                    s.get("provider"),
                ) for s in subtitles
            ])))
    return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from pathlib import Path

from sanic import Sanic

from dopplerr.api import v1
from dopplerr.config import DopplerrConfig
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.periodic_task_manager import PeriodicTaskManager

log = logging.getLogger(__name__)


def listen():
    app = Sanic(__name__, log_config=None)
    app.blueprint(v1.bp)

    DopplerrStatus().healthy = True
    for fi in Path(DopplerrConfig().get_cfg_value("general.frontenddir")).iterdir():
        app.static('/' + fi.name if fi.name != "index.html" else '/', fi.resolve().as_posix())
    app.add_task(PeriodicTaskManager().run())
    app.run(host='0.0.0.0', port=int(DopplerrConfig().get_cfg_value("general.port")))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import asyncio
import os
from datetime import datetime

from apscheduler.schedulers.asyncio import AsyncIOScheduler


def test_apscheduler():
    def tick():
        print('Tick! The time is: %s' % datetime.now())

    async def leave():
        asyncio.get_event_loop().stop()

    scheduler = AsyncIOScheduler()
    scheduler.add_job(tick, 'interval', seconds=1)
    scheduler.add_job(leave, 'interval', seconds=4)
    scheduler.start()
    print('Press Ctrl+{0} to exit'.format('Break' if os.name == 'nt' else 'C'))

    # Execution will block here until Ctrl+C (Ctrl+Break on Windows) is pressed.
    try:
        asyncio.get_event_loop().run_forever()
    except (KeyboardInterrupt, SystemExit):
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from sanic import Blueprint
from sanic.response import json
from sanic.response import text

from dopplerr import DOPPLERR_VERSION
from dopplerr.config import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.downloader import DopplerrDownloader
from dopplerr.plugins.sonarr.task import TaskSonarrOnDownload
from dopplerr.status import DopplerrStatus
from dopplerr.tasks.executors import DopplerrExecutors

log = logging.getLogger(__name__)

bp = Blueprint('api')


@bp.route("/api/v1/recent/events/<num>")
async def recent_events_num(_request, num=10):
    num = int(num)
    if num > 100:
        num = 100
    res = {"events": DopplerrDb().get_recent_events(num)}
    return json(res)


@bp.route("/api/v1/recent/events/")
async def recent_events_10(_request):
    res = {"events": DopplerrDb().get_recent_events(10)}
    return json(res)


@bp.route("/api/v1/recent/fetched/series/<num>")
async def recent_fetched_series_num(_request, num=10):
    num = int(num)
    if num > 100:
        num = 100
    res = {"events": DopplerrDb().get_last_fetched_series(num)}
    return json(res)


@bp.route("/api/v1/notify/sonarr", methods=['POST'])
async def notify_sonarr(request):
    res = await TaskSonarrOnDownload().run(request.json)
    return json(res)


@bp.route("/api/v1/notify", methods=['GET'])
async def notify_not_allowed(_request):
    return text(
        "Method GET not allowed. Use POST with a JSON body with the right format",
        status=405,
    )


@bp.route("/api/v1/health")
async def health(_request):
    res_health = {
        "healthy": DopplerrStatus().healthy,
        "languages": DopplerrConfig().get_cfg_value("subliminal.languages"),
        "mapping": DopplerrConfig().get_cfg_value("general.mapping"),
        "version": DOPPLERR_VERSION,
    }
    return json(res_health)


@bp.route("/api/v1/tasks/status")
async def tasks_status(_request):
    res_health = DopplerrExecutors().status()
    return json(res_health)


@bp.route("/api/v1/version")
async def version(_request):
    return text(DOPPLERR_VERSION)


@bp.route("/api/v1/fullscan")
async def fullscan(request):
    content = request.json
    logging.debug("Fullscan request: %r", content)
    res = await DopplerrDownloader().process_fullscan(content)
    return json(res)


@bp.route("/api/v1/medias/series/")
async def medias_series(_request):
    res = {"medias": DopplerrDb().get_medias_series()}
    return json(res)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from collections import namedtuple

log = logging.getLogger(__name__)

SeriesEpisodeUid = namedtuple('SeriesEpisodeUid', ['tv_db_id', 'season_number', 'episode_number'])

SeriesEpisodeInfo = namedtuple('SeriesEpisodeInfo', [
    'series_episode_uid',
    'episode_title',
    'quality',
    'video_languages',
    'subtitles_languages',
    'media_filename',
    'dirty',
])



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.db import DopplerrDb
from dopplerr.notifications import emit_notifications
from dopplerr.notifications_types.series_subtitles_fetched import SubtitleFetchedNotification
from dopplerr.plugins.sonarr.filter import SonarrFilter
from dopplerr.tasks.download_subtitles import download_missing_subtitles
from dopplerr.tasks.executors import DopplerrExecutors

log = logging.getLogger(__name__)


class DopplerrTask(object):
    def run_in_background(self, task):
        DopplerrExecutors().run_in_background(task)


class TaskSonarrOnDownload(DopplerrTask):
    async def run(self, content):

        logging.debug("Sonarr notification received: %r", content)
        res = await SonarrFilter().filter(content)
        if res.is_unhandled:
            # event has been filtered out
            return res

        self.run_in_background(self.task_sonarr_on_download_background(res))
        return res.to_dict()

    async def task_sonarr_on_download_background(self, res):
        log.debug("Starting task_sonarr_on_download_background")
        res = await download_missing_subtitles(res)
        log.debug("Successful: %r", res.successful)
        if not res.successful:
            return res

        for episode_info in res.sonarr_episode_infos:
            await emit_notifications(SubtitleFetchedNotification(series_episode_info=episode_info))
            DopplerrDb().update_fetched_series_subtitles(
                series_episode_uid=episode_info.series_episode_uid,
                subtitles_languages=episode_info.subtitles_languages,
                dirty=False)
        logging.debug("Background task finished with result: %s", res.to_json())
        return res



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import logging.config
import os
import sys
from pathlib import Path

from babelfish import Language

from dopplerr import DOPPLERR_VERSION
from dopplerr.config import DopplerrConfig
from dopplerr.db import DopplerrDb
from dopplerr.downloader import DopplerrDownloader
from dopplerr.logging_split import setup_logging
from dopplerr.routes import listen
from dopplerr.status import DopplerrStatus

# from dopplerr.logging import setup_logger

log = logging.getLogger(__name__)


def print_list(alist):
    return ",".join(alist)


def list_of_languages(lang_list):
    langs = [s.lower() for s in lang_list.split(',')]
    failed = False
    for l in langs:
        try:
            Language(l)
        except ValueError:
            failed = True
            logging.critical("Invalid language: %r", l)
    if failed:
        sys.exit(2)
    return langs


def main():
    if "-v" in sys.argv or "--general-verbose" in sys.argv:
        default_level = logging.DEBUG
    else:
        default_level = logging.INFO
    debug = default_level is logging.DEBUG
    setup_logging(debug=debug, module_verbose=debug)
    # setup_logger(level=default_level, color=False)
    log.debug("Initializing Dopplerr version %s...", DOPPLERR_VERSION)

    DopplerrConfig().find_configuration_values()

    # setup_logger(
    #     level=(logging.DEBUG
    #            if DopplerrConfig().get_cfg_value("general.verbose") else logging.WARNING),
    #     color=not DopplerrConfig().get_cfg_value("general.no_color"),
    #     logfile=DopplerrConfig().get_cfg_value("general.logfile"))
    debug = DopplerrConfig().get_cfg_value("general.verbose")
    setup_logging(debug=debug, module_verbose=debug)
    log.info("Reset logging format to %s", "verbose"
             if DopplerrConfig().get_cfg_value("general.verbose") else "not verbose")

    log.debug("Applying configuration")
    DopplerrStatus().refresh_from_cfg()

    log.info("Initializing Subtitle DopplerrDownloader Service")

    DopplerrDownloader().initialize_subliminal()
    DopplerrStatus().sqlite_db_path = (
        Path(DopplerrConfig().get_cfg_value("general.configdir")) / "sqlite.db")
    log.debug("SQLite DB: %s", DopplerrStatus().sqlite_db_path.as_posix())
    DopplerrDb().init(DopplerrStatus().sqlite_db_path)
    DopplerrDb().create_tables()

    # change current work dir for subliminal work files
    os.chdir(DopplerrConfig().get_cfg_value("general.configdir"))

    DopplerrDb().insert_event("start", "dopplerr started")

    # main event loop (Asyncio behind)
    listen()

    logging.info("Clean stopping")
    DopplerrDb().insert_event("stop", "dopplerr stopped")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
from pathlib import Path

from peewee import BooleanField
from peewee import CharField
from peewee import DateTimeField
from peewee import ForeignKeyField
from peewee import IntegerField
from peewee import Model
from peewee import PrimaryKeyField
from peewee import TextField
from peewee import Using
from playhouse.sqliteq import SqliteQueueDatabase

from dopplerr.singleton import singleton


# pylint: disable=invalid-name
class Events(Model):
    timestamp = DateTimeField(default=datetime.datetime.now)
    type = CharField()
    message = TextField()


class SeriesMedias(Model):
    id = PrimaryKeyField()
    added_timestamp = DateTimeField(default=datetime.datetime.now)
    series_title = CharField(null=True)
    tv_db_id = IntegerField(null=False)
    season_number = IntegerField(null=False)
    episode_number = IntegerField(null=False)
    episode_title = CharField(null=True)
    quality = CharField(null=True)
    video_languages = CharField(null=True)
    dirty = BooleanField(default=True)
    media_filename = TextField(null=True)


class SeriesSubtitles(Model):
    added_timestamp = DateTimeField(default=datetime.datetime.now)
    series_media = ForeignKeyField(SeriesMedias, related_name='subtitles')
    language = CharField()


# pylint: enable=invalid-name


@singleton
class DopplerrDb(object):
    def __init__(self):
        self.__sqlite_db_path = None
        self.__conn = None
        self.__database = None

    @property
    def database(self):
        if not self.__database:
            self.__database = SqliteQueueDatabase(
                self.__sqlite_db_path, pragmas=(('foreign_keys', 'on'), ))
        return self.__database

    @property
    def conn(self):
        return self.database.get_conn()

    def init(self, sqlite_db_path: Path):
        self.__sqlite_db_path = sqlite_db_path.as_posix()

    def create_tables(self):
        self.database.create_table(Events, safe=True)
        self.database.create_table(SeriesSubtitles, safe=True)
        self.database.create_table(SeriesMedias, safe=True)

    def insert_event(self, thetype: str, message: str):
        with Using(self.database, [Events], with_transaction=False):
            Events.create(type=thetype, message=message)

    def get_recent_events(self, limit: int):
        with Using(self.database, [Events], with_transaction=False):
            events = (Events.select().limit(limit).order_by(Events.timestamp.desc()).execute())
            return [{
                "timestamp": e.timestamp.strftime('%Y-%m-%dT%H:%M:%S'),
                "type": e.type,
                "message": e.message
            } for e in events]

    def update_series_media(self,
                            series_title,
                            tv_db_id,
                            season_number,
                            episode_number,
                            episode_title,
                            quality,
                            video_languages,
                            media_filename,
                            dirty=True):
        with Using(self.database, [SeriesMedias], with_transaction=False):
            media, _ = SeriesMedias.get_or_create(
                tv_db_id=tv_db_id, season_number=season_number, episode_number=episode_number)
            media.series_title = series_title
            media.episode_title = episode_title
            media.quality = quality
            media.video_languages = video_languages
            media.dirty = dirty
            media.media_filename = media_filename
            media.save()

    def update_fetched_series_subtitles(self, series_episode_uid, subtitles_languages, dirty=True):
        with Using(self.database, [SeriesMedias, SeriesSubtitles], with_transaction=False):
            media = (SeriesMedias.select().where(
                SeriesMedias.tv_db_id == series_episode_uid.tv_db_id,
                SeriesMedias.season_number == series_episode_uid.season_number,
                SeriesMedias.episode_number == series_episode_uid.episode_number))
            for lang in subtitles_languages:
                SeriesSubtitles.get_or_create(
                    series_media=media,
                    language=lang,
                )
            for m in media:
                m.dirty = dirty
                m.save()

    def get_last_fetched_series(self, limit: int):
        with Using(self.database, [SeriesMedias, SeriesSubtitles], with_transaction=False):
            events = (SeriesSubtitles.select().order_by(
                SeriesSubtitles.added_timestamp.desc()).limit(limit).execute())
            return [{
                "added_timestamp": e.added_timestamp.strftime('%Y-%m-%dT%H:%M:%S'),
                "series_title": e.series_media.series_title,
                "season_number": e.series_media.season_number,
                "episode_number": e.series_media.episode_number,
                "episode_title": e.series_media.episode_title,
                "quality": e.series_media.quality,
                "video_languages": e.series_media.video_languages,
                "subtitle_language": e.language,
            } for e in events]

    def get_medias_series(self):
        with Using(self.database, [SeriesMedias, SeriesSubtitles], with_transaction=False):
            limit = 100
            medias = (SeriesMedias.select().order_by(
                SeriesMedias.series_title.desc()).limit(limit).execute())
            return [{
                "added_timestamp": med.added_timestamp.strftime('%Y-%m-%dT%H:%M:%S'),
                "series_title": med.series_title,
                "season_number": med.season_number,
                "episode_number": med.episode_number,
                "episode_title": med.episode_title,
                "quality": med.quality,
                "video_languages": med.video_languages,
                "subtitle_languages": sorted([s.language for s in med.subtitles]),
                "dirty": med.dirty,
            } for med in medias]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
from pathlib import PosixPath

import pkg_resources

from cfgtree.cfgtree import EnvironmentConfig
from cfgtree.storages import JsonFileConfigStorage
from cfgtree.types import BoolCfg
from cfgtree.types import ConfigFileCfg
from cfgtree.types import DirNameCfg
from cfgtree.types import IntCfg
from cfgtree.types import ListOfStringCfg
from cfgtree.types import MultiChoiceCfg
from cfgtree.types import PasswordCfg
from cfgtree.types import StringCfg
from cfgtree.types import UserCfg
from dopplerr.singleton import singleton

log = logging.getLogger(__name__)
DEFAULT_PORT = 8086


def _find_frontend_data():
    installed_data_frontend = pkg_resources.resource_filename(__name__, 'frontend')
    if PosixPath(installed_data_frontend).exists():
        log.debug("Found local frontend path: %s", installed_data_frontend)
        return installed_data_frontend
    setup_py = pkg_resources.resource_filename(__name__, "main.py")
    dev_env_frontend_dist = PosixPath(setup_py).parent.parent / "frontend" / "dist"
    if dev_env_frontend_dist.exists():
        log.debug("Found dev local frontend path: %s", dev_env_frontend_dist)
        return str(dev_env_frontend_dist)
    return None


class DopplerrJsonConfigFile(JsonFileConfigStorage):
    json_configstorage_environ_var_name = "DOPPLERR_COMMON_CONFIG_FILE"
    json_configstorage_long_param_name = "--configfile"
    json_configstorage_short_param_name = "-g"
    json_configstorage_default_filename = "config.json"


@singleton
class DopplerrConfig(EnvironmentConfig):

    environ_var_prefix = "DOPPLERR_"
    config_storage = DopplerrJsonConfigFile()

    cfgtree = {
        "configfile": ConfigFileCfg(l="--configfile", h="Config directory"),
        "general": {
            "basedir":
                DirNameCfg(s="-b", d=os.getcwd(), h='Base directory'),
            "configdir":
                DirNameCfg(s="-c", d=os.getcwd(), h="Config directory"),
            "appdir":
                DirNameCfg(s="-a", d=os.getcwd(), h="App directory"),
            "frontenddir":
                DirNameCfg(s="-f", d=_find_frontend_data(), r=True, h="Frontend directory"),
            "verbose":
                BoolCfg(s='-v', h='Enable verbose output'),
            "logfile":
                StringCfg(s="-l", h='Output log to file'),
            "mapping":
                ListOfStringCfg(
                    s='-m',
                    h=("Map root folder of tv/anime/movie to another name.\n"
                       "Ex: series are mounted on a docker image as /tv but \n"
                       "on the other system it is under /video/Series. In this \n"
                       "case use '--basedir /video --mapping tv=Series,movies=Movies'")),
            "port":
                IntCfg(s='-p', d=DEFAULT_PORT, h='The port to listen on'),
            "no_color":
                BoolCfg(h="Disable color in logs"),
        },
        "subliminal": {
            "languages": ListOfStringCfg(),
            "addic7ed": {
                "enabled": BoolCfg(h="Enable addic7ed"),
                "user": UserCfg(h="addic7ed username"),
                "password": PasswordCfg(h="addic7ed password"),
            },
            "legendastv": {
                "enabled": BoolCfg(h="Enable legendastv"),
                "user": UserCfg(h="legendastv username"),
                "password": PasswordCfg(h="legendastv password"),
            },
            "opensubtitles": {
                "enabled": BoolCfg(h="Enable opensubtitles"),
                "user": UserCfg(h="opensubtitles username"),
                "password": PasswordCfg(h="opensubtitles password"),
            },
            "subscenter": {
                "enabled": BoolCfg(h="Enable subscenter"),
                "user": UserCfg(h="subscenter username"),
                "password": PasswordCfg(h="subscenter password"),
            },
        },
        "notifications": {
            "pushover": {
                "enabled":
                    BoolCfg(h="Enable pushover"),
                "user":
                    UserCfg(h="pushover username"),
                "token":
                    PasswordCfg(h="pushover password"),
                "registered_notifications":
                    MultiChoiceCfg(h="Registerd Notifications", choices=["fetched"], d=["fetched"]),
            }
        }
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.notifications.pushover import NotificationPushOver
from dopplerr.config import DopplerrConfig

log = logging.getLogger(__name__)


async def emit_notifications(notification):
    log.debug("Emiting notification: [%s] %s - %s", notification.notification_type,
              notification.notification_title, notification.one_liner)
    if DopplerrConfig().get_cfg_value("notifications.pushover.enabled"):
        log.debug("Emiting pushover with user %s",
                  DopplerrConfig().get_cfg_value("notifications.pushover.user"))
        po = NotificationPushOver(
            DopplerrConfig().get_cfg_value("notifications.pushover.token"),
            DopplerrConfig().get_cfg_value("notifications.pushover.user"),
            DopplerrConfig().get_cfg_value("notifications.pushover.registered_notifications"),
        )
        await po.emit(notification.notification_type, notification.notification_title,
                      notification.one_liner)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import glob
import sys


def recursive_iglob_fallback(path):
    try:
        # pylint: disable=import-error
        import glob2
        # pylint: enable=import-error
    except Exception:
        raise Exception("glob2 not installed on your environment !")
    return glob2.iglob(path, with_matches=True)


def recursive_iglob(path):
    if sys.version_info < (3, 5):
        return recursive_iglob_fallback(path)
    return glob.iglob(path, recursive=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import asyncio
import concurrent
import functools
import logging

from dopplerr.singleton import singleton

log = logging.getLogger(__name__)


class _ExecutorsBase(object):
    parallel_executors = 1

    def __init__(self):
        self.executors = concurrent.futures.ThreadPoolExecutor(max_workers=self.parallel_executors)

    async def run_in_thread(self, func, *args, **kwargs):
        event_loop = asyncio.get_event_loop()
        res = await event_loop.run_in_executor(self.executors,
                                               functools.partial(func, *args, **kwargs))
        return res


class _SubliminalExecutors(_ExecutorsBase):
    parallel_executors = 1


@singleton
class DopplerrExecutors(object):
    background_tasks = 0

    def __init__(self):
        self.subliminal_executors = _SubliminalExecutors()

    def run_in_background(self, task):
        async def wrap_task(task):
            self.background_tasks += 1
            res = await task
            self.background_tasks -= 1
            return res

        loop = asyncio.get_event_loop()
        loop.create_task(wrap_task(task))

    def status(self):
        return {
            'background_tasks': self.background_tasks,
        }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
import sys
from functools import partial
from io import StringIO
from subprocess import PIPE
from subprocess import check_output

# absolute_import ensure "import logging" will not import the logging package from the
# current module but the system wide package

log = logging.getLogger(__name__)

# flake8: noqa
# pylint: disable=too-many-locals,bare-except


def setup_logging(split=True,
                  unbuffered=False,
                  debug=False,
                  module_verbose=False,
                  fmt=None,
                  force_no_tty=None):
    '''
    :param split: preserve log headers for multiline strings, and adapt to the current terminal
                  witdh if applicable
    :param unbuffered: ask the stdout to avoid buffering to send logs as fast as possible to the
                       caller script, only when not inside a TTY
    :param debug: enable debug level
    :param module_verbose: add module (file, line number) to each log line
    :param fmt: overwrite the formatter
    :param force_no_tty: force the usage of the no TTY formatter
    '''
    try:
        # Note: colorlog is not a 'requirement' of slavescripts (and shouldn't!), so it might not be
        # installed. In this case, normal logging formatter is used. But if available, let's print
        # plentful of colors!
        #
        # To install it, simply do:
        #
        #    pip install colorlog
        from colorlog import ColoredFormatter
    except:
        ColoredFormatter = None  # pylint: disable=invalid-name

    # this logging config will output debug and info to stdout and warning, error, and critical to
    # stderr, so that buildbot can color them
    # logs with level > will go to stderr, with level < will go to stdout
    stderr_threshold = logging.WARNING
    # logs with level < will be hidden
    stdout_level = logging.DEBUG

    class InfoFilter(logging.Filter):
        def filter(self, record):
            return record.levelno < stderr_threshold

    class SplitFormatterMixin(object):
        '''
        Magic Formatter that gracefully handle multiline logs, ie it split the multiline string
        and add each log independently.

        This handles two spliting:

        - if the terminal width is found, try to fit the output in the terminal by splitting the
          string arbitrarily
        - if the record message is multiline, split it and print prefix (loglevel, date,...)
          independently

        Note this only impacts the display to stdout handler, not the record storage, so
        if these logs are sent to another handler (for instance to splunk handler), the log
        will not be split
        '''

        def format(self, record):
            record.message = record.getMessage()
            record.msg = record.message
            if record.args:
                record.args = ()
            if self.usesTime():
                record.asctime = self.formatTime(record, self.datefmt)
            multiline_message = record.message
            formatted_lines = []
            if not self.split:
                s = self._get_formated_line(record)
                formatted_lines.append(s)
            else:
                for line in multiline_message.split("\n"):
                    if self.term_width < 1:
                        sublines = [line]
                    else:
                        sublines = [
                            l for l in iter(partial(StringIO(line).read, int(self.term_width)), '')
                        ]
                    if not sublines:
                        # If the string is empty, still add an empty record in the output
                        record.message = ""
                        record.msg = ""
                        s = self._get_formated_line(record)
                        formatted_lines.append(s)
                    else:
                        for subline in sublines:
                            record.message = subline
                            record.msg = subline
                            s = self._get_formated_line(record)
                            formatted_lines.append(s)
            return "\n".join(formatted_lines)

        def _get_formated_line(self, record):
            s = self._get_record_with_fmt(record)
            if record.exc_info:
                # Cache the traceback text to avoid converting it multiple times
                # (it's constant anyway)
                if not record.exc_text:
                    record.exc_text = self.formatException(record.exc_info)
            if record.exc_text is not None:
                if s and s[-1:] != "\n":
                    s = s + "\n"
                try:
                    s = s + record.exc_text
                except UnicodeError:
                    # Sometimes filenames have non-ASCII chars, which can lead
                    # to errors when s is Unicode and record.exc_text is str
                    # See issue 8924.
                    # We also use replace for when there are multiple
                    # encodings, e.g. UTF-8 for the filesystem and latin-1
                    # for a script. See issue 13232.
                    s = s + record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
            return s

    class SplitFormatter(SplitFormatterMixin, logging.Formatter):
        def __init__(self, fmt=None, datefmt=None, term_width=-1, split=True):
            super(SplitFormatter, self).__init__(fmt, datefmt)
            self.term_width = term_width
            self.split = split

        def _get_record_with_fmt(self, record):
            return self._fmt % record.__dict__

    if ColoredFormatter:

        class SplitColorFormatter(SplitFormatterMixin, ColoredFormatter):
            def __init__(self, fmt=None, datefmt=None, term_width=-1, split=True, log_colors=None):
                super(SplitColorFormatter, self).__init__(fmt, datefmt, log_colors=log_colors)
                self.term_width = term_width
                self.split = split

            def _get_record_with_fmt(self, record):
                return ColoredFormatter.format(self, record)

    align_level_width = 8
    extra_char_width = 3
    if fmt:
        ColoredFormatter = None  # pylint: disable=invalid-name
        fmt_nocolor_str = fmt
    else:
        if module_verbose:
            fmt_module_nocolor_str = "[%(name)-37.37s] "
            fmt_module_color_str = ("[%(yellow)s%(name)-37.37s%(reset)s] ")
        else:
            fmt_module_nocolor_str = ""
            fmt_module_color_str = ""
        fmt_debug_nocolor_str = (
            '%(asctime)19.19ss ' + fmt_module_nocolor_str + '%(levelname)7s: %(message)s')
        fmt_debug_color_str = (
            '%(blue)s%(asctime)19.19s%(reset)s ' + fmt_module_color_str +
            '%(log_color)s%(levelname)-7s%(reset)s | ' + '%(log_color)s%(message)s%(reset)s')

        fmt_simple_nocolor_str = '%(levelname)-7s - ' + fmt_module_nocolor_str + '%(message)s'
        fmt_simple_color_str = (
            '%(log_color)s' + fmt_module_color_str + '[%(levelname).1s]%(reset)s %(log_color)s'
            '%(message)s%(reset)s')

        if debug:
            fmt_color_str = fmt_debug_color_str
            fmt_nocolor_str = fmt_debug_nocolor_str
        else:
            fmt_color_str = fmt_simple_color_str
            fmt_nocolor_str = fmt_simple_nocolor_str

    if debug:
        default_level = logging.DEBUG
    else:
        default_level = logging.INFO

    logging.basicConfig(stream=sys.stdout, level=default_level, format=fmt_nocolor_str)

    date_fmt_string = None
    # Replace the default formatter because it is buggy
    root = logging.getLogger()
    # Do *not* replace the formatter in quiet mode since we want to bare output

    term_width = -1
    if split:
        # Try to retrieve the terminal width:
        term_width = -1
        try:
            # redirecting the stderr to an unused PIPE, to avoid the following issue:
            #     'stty: standard input: Invalid argument'
            # when executed from a step ShellCommand
            data = check_output(['stty', 'size'], stderr=PIPE)
            _, columns = data.split()
            if columns > 0:
                term_width = columns
        except:
            pass
        term_width = (int(term_width) - align_level_width - extra_char_width)
    # disable terminal spliting for the moment
    term_width = -1

    if force_no_tty:
        is_tty = False
    else:
        is_tty = sys.stdout.isatty()

    if not is_tty:
        # We are in Buildbot or in a pipe, don't use colorlog!
        ColoredFormatter = None  # pylint: disable=invalid-name
    if not ColoredFormatter:
        formatter = SplitFormatter(fmt_nocolor_str, term_width=term_width, split=True)
    else:
        formatter = SplitColorFormatter(
            fmt_color_str,
            date_fmt_string,
            term_width=term_width,
            split=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bg_white',
            })

    if unbuffered and not is_tty:
        root.handlers = []
        # Make stdout unbuffered, so that we get output asap to buildbot log
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 0)
        h1 = logging.StreamHandler(sys.stdout)
        h1.setLevel(stdout_level)
        h1.setFormatter(formatter)
        h1.addFilter(InfoFilter())
        h2 = logging.StreamHandler(sys.stderr)
        h2.setFormatter(formatter)
        h2.setLevel(stderr_threshold)

        # Don't print useless info about "New connection" + security warnings
        logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.ERROR)

        logging.getLogger().addHandler(h1)
        logging.getLogger().addHandler(h2)
        logging.getLogger().setLevel(logging.NOTSET)
    else:
        root.handlers[0].setFormatter(formatter)
    return root.handlers[0]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import asyncio
import logging

from dopplerr.singleton import singleton

log = logging.getLogger(__name__)


@singleton
class PeriodicTaskManager(object):
    running = True

    async def run(self):
        while self.running:
            await asyncio.sleep(2)
            if not self.running:
                return
            log.debug("PERIODIC TASK MANAGER !!!!")

    def stop(self):
        self.running = False



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import os

_UNDEFINED = object()


class _CfgBase(object):

    default = None
    name = None
    xpath = None
    arg_type = None
    environ_var_prefix = None
    ignore_in_cfg = False

    def __init__(self, l=None, s=None, h=None, r=False, d=_UNDEFINED):
        # Note: self.name should come later by EnvironmentConfig._inject_names()
        self.short_param = s
        self.help_str = h
        self.required = r
        self.forced_long_param = l
        if d != _UNDEFINED:
            self.default = d
        self.value = self.default

    def set_value(self, value):
        """
        Setter method used in set_node_by_xpath
        """
        self.value = value

    @property
    def environ_var_name(self):
        return self.environ_var_prefix + self.cmd_line_name.upper()

    def get_cmd_line_params(self):
        a = []
        if self.short_param:
            a.append(self.short_param)
        if self.name:
            a.append(self.long_param)
        return a

    @property
    def _environ_var_value(self):
        return os.environ.get(self.environ_var_name, _UNDEFINED)

    def read_environ_var(self):
        return str(self._environ_var_value)

    @property
    def long_param(self):
        if self.forced_long_param:
            return self.forced_long_param
        if not self.xpath:
            return "--" + self.name.lower().replace("_", "-")
        return "--" + self.xpath.replace('.', '-').replace('_', '-')

    @property
    def cmd_line_name(self):
        return self.xpath.lower().replace("-", "_").replace(".", "_")

    @property
    def action(self):
        return 'store'

    @property
    def n_args(self):
        return None

    @property
    def safe_value(self):
        """
        string that can be outputed to logs
        """
        return self.value

    @property
    def cfgfile_value(self):
        """
        value as it should be saved in config file
        """
        return self.value if self.value is not None else ""

    @property
    def metavar(self):
        return self.name.upper()


class StringCfg(_CfgBase):
    default = ""

    def read_environ_var(self):
        return str(self._environ_var_value)


class ListOfStringCfg(_CfgBase):
    """
    Comma separated list of string (1 argument)
    """

    def __init__(self, *args, **kwargs):
        self.default = []
        super(ListOfStringCfg, self).__init__(*args, **kwargs)

    def read_environ_var(self):
        ls = self._environ_var_value
        return ls.split(",")

    @property
    def cfgfile_value(self):
        """
        value as it should be saved in config file
        """
        return ",".join(self.value)

    @staticmethod
    def arg_type(string):
        return string.split(",")


class IntCfg(_CfgBase):
    default = 0

    def read_environ_var(self):
        return int(self._environ_var_value)


class UserCfg(StringCfg):
    @property
    def user(self):
        return self.value


class PasswordCfg(StringCfg):
    @property
    def password(self):
        return self.value

    @property
    def safe_value(self):
        """
        Hide password in logs
        """
        return "*" * len(self.value)


class DirNameCfg(StringCfg):
    default = None


class ConfigFileCfg(StringCfg):
    default = None
    ignore_in_cfg = True


class BoolCfg(_CfgBase):
    default = False

    def read_environ_var(self):
        e = os.environ.get(self.environ_var_name)
        return bool(e)

    @property
    def action(self):
        return 'store_true'

    @property
    def metavar(self):
        return None


class MultiChoiceCfg(ListOfStringCfg):
    def __init__(self, choices=None, *args, **kwargs):
        super(MultiChoiceCfg, self).__init__(*args, **kwargs)
        self.choices = choices

    def arg_type(self, string):
        items = string.split(",")
        for item in items:
            if item not in self.choices:
                raise argparse.ArgumentTypeError("{!r} not in available choise: {}".format(
                    item, ", ".join(self.choices)))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from unittest import TestCase

from cfgtree.dictxpath import delete_node_by_xpath
from cfgtree.dictxpath import get_node_by_xpath
from cfgtree.dictxpath import set_node_by_xpath


class DictXpathTests(TestCase):
    def test_get_node_by_path(self):
        mapping = {'level1': {'level2': {'level3': 42}}}
        expected = 42
        actual = get_node_by_xpath(mapping, 'level1.level2.level3')
        self.assertEqual(expected, actual)
        self.assertRaisesRegex(
            KeyError, (r"Invalid 'level1.unknown' selector: 'unknown' doesn't match anything. "
                       r"Available keys: \['level2'\]"), get_node_by_xpath, mapping,
            'level1.unknown')
        actual = get_node_by_xpath(
            mapping, 'level1.unknown', default="default value", ignore_errors=True)
        self.assertEqual(actual, "default value")

    def test_get_node_by_path_empty_mapping(self):
        empty_map = {}
        self.assertRaises(KeyError, get_node_by_xpath, empty_map, 'level1')
        self.assertRaises(KeyError, get_node_by_xpath, empty_map, 'level1.other')
        self.assertEqual(
            get_node_by_xpath(empty_map, 'level1', default="default value", ignore_errors=True),
            "default value")
        self.assertEqual(
            get_node_by_xpath(empty_map, '', default="default value", ignore_errors=True),
            "default value")

    def test_get_node_by_path_invalid_first_level(self):
        mapping = {"level1": {}}
        self.assertRaises(KeyError, get_node_by_xpath, mapping, 'invalid_level1')
        self.assertEqual(
            get_node_by_xpath(
                mapping, 'invalid_level1', default="default value", ignore_errors=True),
            "default value")

    def test_get_node_by_path_mapping_not_dict(self):
        mapping_no_a_dict = "simple string!"
        self.assertEqual(
            get_node_by_xpath(mapping_no_a_dict, '', default="default value", ignore_errors=True),
            "default value")
        self.assertRaises(
            KeyError,
            get_node_by_xpath,
            mapping_no_a_dict,
            'level1.unknown',
            ignore_errors=False,
        )

    def test_get_node_by_path_incomplete_mapping(self):
        mapping = {'level1': {'level2': {}}}
        self.assertRaises(KeyError, get_node_by_xpath, mapping, 'level1.level2.level3')
        self.assertRaises(KeyError, get_node_by_xpath, mapping, 'level1.unknown')
        # test with tailing "." at the end of path
        self.assertRaises(KeyError, get_node_by_xpath, mapping, 'level1.')
        self.assertEqual(
            get_node_by_xpath(
                mapping, 'level1.level2', default="default value", ignore_errors=True), {})
        self.assertEqual(
            get_node_by_xpath(
                mapping, 'level1.unknown', default="default value", ignore_errors=True),
            "default value")
        # test with tailing "." at the end of path
        self.assertEqual(
            get_node_by_xpath(mapping, 'level1.', default="default value", ignore_errors=True),
            "default value")

    def test_set_node_by_path(self):
        mapping = {'level1': {'level2': {'level3': None}}}
        expected = 42
        set_node_by_xpath(mapping, 'level1.level2.level3', expected)
        actual = get_node_by_xpath(mapping, 'level1.level2.level3')
        self.assertEqual(expected, actual)
        set_node_by_xpath(mapping, 'level1.unknown', expected)
        actual = get_node_by_xpath(mapping, 'level1.unknown')
        self.assertEqual(expected, actual)
        self.assertRaises(KeyError, set_node_by_xpath, mapping, 'level1.invalid.level3', '')
        expected = 'extended'
        set_node_by_xpath(mapping, 'level1.missing.level3', expected, extend=True)
        actual = get_node_by_xpath(mapping, 'level1.missing.level3')
        self.assertEqual(expected, actual)

    def test_delete_node_by_path(self):
        mapping = {'level1': {'level2': {'level3': 42}}}
        expected = 42
        actual = delete_node_by_xpath(mapping, 'level1.level2.level3')
        self.assertEqual(expected, actual)
        self.assertFalse('level3' in mapping['level1']['level2'])
        self.assertRaises(KeyError, delete_node_by_xpath, mapping, 'level1.invalid.level3')
        self.assertIsNone(
            delete_node_by_xpath(mapping, 'level1.invalid.level3', ignore_errors=True))

    def test_get_node_by_path_with_list_selector(self):
        mapping = {'level1': {'level_2_is_a_list': ['item1', 'item2']}}
        actual = get_node_by_xpath(
            mapping, 'level1.level_2_is_a_list[1]', handle_list_selector=True)
        self.assertEqual('item2', actual)

    def test_get_node_w_lst_selctr_sub_list(self):
        mapping = {
            'level1': {
                'level_2_is_a_list': [
                    {
                        'item1': {
                            'k1': 'v1'
                        }
                    },
                    {
                        'item2': {
                            'k2': 'v2'
                        }
                    },
                ]
            }
        }
        actual = get_node_by_xpath(
            mapping, 'level1.level_2_is_a_list[1].item2.k2', handle_list_selector=True)
        self.assertEqual('v2', actual)

    def test_get_w_bad_lst_selctr_n_default_val(self):
        mapping = {'level1': {'level_2_is_a_list': ['item1', 'item2']}}
        actual = get_node_by_xpath(
            mapping, 'level1.level_2_is_a_list[1]', handle_list_selector=True, default="N/A")
        self.assertEqual('item2', actual)
        self.assertRaisesRegex(
            KeyError,
            (r"Invalid \'level1.level_2_is_a_list\[99\]\' selector: "
             r"item index \'99\' of \'level_2_is_a_list\' is outside of the list boundaries. "
             r"Length is: 2"),
            get_node_by_xpath,
            mapping,
            'level1.level_2_is_a_list[99]',
            handle_list_selector=True,
            default="N/A")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

log = logging.getLogger(__name__)


class _NotificationTypeBase(object):
    pass


class _SeriesNotificationBase(_NotificationTypeBase):
    notification_type = NotImplementedError
    notification_title = NotImplementedError
    series_episode_info = None

    def __init__(self, series_episode_info):
        self.series_episode_info = series_episode_info



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import tempfile
from logging.handlers import RotatingFileHandler
from pathlib import Path

# pylint: disable=import-error
try:
    import colorlog
except Exception:
    colorlog = None
# pylint: enable=import-error


def temp_dir(name, root=None):
    root = root or tempfile.gettempdir()
    directory = Path(root) / name
    directory.mkdir(exist_ok=True)
    return directory


class ColorCmd(object):
    LOG_COLOR = "%(log_color)s"
    RESET = "%(reset)s"
    MSG_COLOR = "%(message_log_color)s"


class StdOutLogChunks(object):
    DATE = "%(asctime)19s"
    LEVEL = "[ %(levelname)-8s ]"
    NAME = "[ %(name)-25s ]"
    MESSAGE = "%(message)s"


class StdOutColorLogChunks(object):
    DATE = StdOutLogChunks.DATE
    LEVEL = ColorCmd.LOG_COLOR + "%(white)s[ %(reset)s%(levelname)-8s%(white)s ]%(reset)s"
    NAME = ColorCmd.LOG_COLOR + "%(white)s[ %(reset)s%(name)-25s%(white)s ]%(reset)s"
    MESSAGE = ColorCmd.MSG_COLOR + StdOutLogChunks.MESSAGE + ColorCmd.RESET


class StdOutLog(object):
    color_fmt = "{date} {name} {level} {message}".format(
        date=StdOutColorLogChunks.DATE,
        level=StdOutColorLogChunks.LEVEL,
        message=StdOutColorLogChunks.MESSAGE,
        name=StdOutColorLogChunks.NAME,
    )
    no_color_fmt = "{date} {name} {level} {message}".format(
        date=StdOutLogChunks.DATE,
        level=StdOutLogChunks.LEVEL,
        message=StdOutLogChunks.MESSAGE,
        name=StdOutLogChunks.NAME,
    )


class DefaultFileFmtChunks(object):
    DATE = "%(asctime)s"
    LEVELNAME = "%(levelname)-8s"
    FILE_LINE = "%(pathname)s:%(lineno)s"
    MESSAGE = "%(message)s"


class DefaultFileFmt(object):
    fmt = " :: ".join([
        DefaultFileFmtChunks.DATE,
        DefaultFileFmtChunks.LEVELNAME,
        DefaultFileFmtChunks.FILE_LINE,
        DefaultFileFmtChunks.MESSAGE,
    ])


g_file_handler = None
g_stream_handler = None


def setup_logger(
        color=True,
        level=logging.INFO,
        logfile=None,
        file_fmt=DefaultFileFmt.fmt,
        stdout_color_fmt=StdOutLog.color_fmt,
        stdout_fmt=StdOutLog.no_color_fmt,
):
    """
    Simple yet efficient logger configuration utility.

    Typically, you call this method as soon as possible in your application, so you have logs
    on your terminal. Then you recall it with options, depending on your application configuration
    (ex: verbose, output to log file, ...)

    Usage:
    ```
    def main():
        setup_logger()
        ...
        # read command line argument
        verbose = opt/argparse.verbose
        setup_logger(level=logging.DEBUG is verbose else logging.INFO)
    ```
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    global g_file_handler
    global g_stream_handler

    if logfile:
        file_formatter = logging.Formatter(file_fmt)
        file_handler = RotatingFileHandler(str(logfile), 'a', 5 * 1024 * 1024, 1)
        file_handler.setFormatter(file_formatter)
        if g_file_handler:
            root_logger.removeHandler(g_file_handler)
        root_logger.addHandler(file_handler)
        g_file_handler = file_handler

    if color and colorlog is not None:
        format_str = stdout_color_fmt
        stream_handler = colorlog.StreamHandler()
        colored_formatter = colorlog.ColoredFormatter(
            format_str,
            log_colors={
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bg_white',
                'DEBUG': 'cyan',
            },
            secondary_log_colors={
                'message': {
                    'ERROR': 'red',
                    'CRITICAL': 'red,bg_white',
                    'DEBUG': 'cyan',
                }
            })
        stream_handler.setFormatter(colored_formatter)
    else:
        format_str = stdout_fmt
        stream_handler = logging.StreamHandler()
        stream_formatter = logging.Formatter(format_str)
        stream_handler.setFormatter(stream_formatter)
    if g_stream_handler:
        root_logger.removeHandler(g_stream_handler)
    root_logger.addHandler(stream_handler)
    g_stream_handler = stream_handler



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.response import Response

log = logging.getLogger(__name__)


class SonarrOnDownloadResponse(Response):
    def __init__(self, *args, **kwargs):
        super(SonarrOnDownloadResponse, self).__init__(*args, **kwargs)
        self.request_type = "sonarr"
        self.request_event = "on download"

    @property
    def subtitles(self):
        return self.res.get("result", {}).get("subtitles", None)

    @subtitles.setter
    def subtitles(self, subtitles):
        self.res.setdefault("result", {})["subtitles"] = subtitles

    @property
    def candidates(self):
        return self.res.setdefault("candidates", [])

    @candidates.setter
    def candidates(self, candidates):
        self.res.candidates = candidates

    @property
    def sonarr_episode_infos(self):
        candidate = self.candidates[0]
        subtitles = self.subtitles
        return [{
            "series_title": candidate.get("series_title"),
            "season_number": candidate.get("season_number"),
            "tv_db_id": candidate.get("tv_db_id"),
            "episode_number": candidate.get("episode_number"),
            "episode_title": candidate.get("episode_title"),
            "quality": candidate.get("quality"),
            "video_languages": candidate.get("video_languages", "???"),
            "subtitles_languages": [s["language"] for s in subtitles],
        }]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import asyncio
import time

import pytest

from dopplerr.tasks.executors import DopplerrExecutors
from dopplerr.tasks.executors import _ExecutorsBase


class OnlyOneExecution(_ExecutorsBase):
    parallel_executors = 1


@pytest.mark.asyncio
async def test_executors():
    print("")
    de = DopplerrExecutors()

    def mkprefix(i):
        return " " * 15 + " " * 45 * i

    def long_sync_task(i):
        # all tasks should run inside the SAME thread
        prefix = mkprefix(1)
        print(prefix + "long  sync task {}: begin".format(i))
        for _ in range(0, i * 10):
            print(prefix + "long  sync task {}: working".format(i))
            # note: time.sleep does NOT block current thread, so other "blocking" task
            # might actually overlap their executions
            time.sleep(0.1)
        print(prefix + "long  sync task {}: end".format(i))

    async def long_async_task(i):
        print("long async task {}: begin".format(i))
        await OnlyOneExecution().run_in_thread(long_sync_task, i)
        print("long async task {}: more work after results from run in thread task".format(i))
        await asyncio.sleep(1)
        print("long async task {}: end".format(i))

    print("event loop thread" + " " * 40 + "worker thread")
    de.run_in_background(long_async_task(1))
    de.run_in_background(long_async_task(2))
    de.run_in_background(long_async_task(3))
    de.run_in_background(long_async_task(4))
    print("waiting 5s")
    for _ in range(0, 60):
        print("-- running long async tasks: {} --".format(de.status()['background_tasks']))
        await asyncio.sleep(0.1)
    print("end of unittest")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

from dopplerr.notifications_types._base import _SeriesNotificationBase

log = logging.getLogger(__name__)


class SeriesMediaRefreshedNotification(_SeriesNotificationBase):
    notification_type = "refresh"
    notification_title = "Episode Information Refreshed"

    @property
    def one_liner(self):
        return ("{e.series_title} - {e.season_number}x{e.episode_number} - "
                "{e.episode_title} [{e.quality}] - Lang: {video_languages}".format(
                    e=self.series_episode_info,
                    video_languages=",".join(self.series_episode_info.video_languages),
                ))
